/**
 * sample for reproducing the issue of being unable to shoot continuously with gphoto2.
 *
 * dependency: libgphoto2
 */

#include <filesystem>
#include <iostream>
#include <fcntl.h> // C file I/O

#include <boost/algorithm/string.hpp>
#include <boost/thread/mutex.hpp>

#include <gphoto2/gphoto2-camera.h>
#include <gphoto2/gphoto2-context.h>
#include <gphoto2/gphoto2-setting.h>
#include <gphoto2/gphoto2-filesys.h>

// from: https://stackoverflow.com/questions/59883881/easiest-way-to-print-timestamp-to-ostream
class SimpleLogger {
private:
  std::ostream &stream;

  void print_time() {
    std::time_t t = std::time(nullptr);
    stream << "[" << std::put_time(std::localtime(&t), "%F %T %Z") << "] ";
  }
public:
  //Maybe also take options for how to log?
  SimpleLogger(std::ostream &stream) : stream(stream) { }

  template <typename T>
  std::ostream &operator<<(const T &thing)  {
    print_time();
    return stream << thing;
  }
};

SimpleLogger logger_(std::cerr);

Camera* camera_;
GPContext *context_;

boost::mutex photo_mutex_;
auto download_image_(1);
auto trigger_wait_timeout_(50.0);
std::filesystem::path download_path_ = "/tmp";
auto delete_image_after_download_(false);

/*
 * just triggers capture
 * ref: https://cpp.hotexamples.com/examples/-/-/gp_camera_wait_for_event/cpp-gp_camera_wait_for_event-function-examples.html
 */

bool do_trigger(std::filesystem::path &path, const int wait_timeout, const CameraEventType wait_for = GP_EVENT_CAPTURE_COMPLETE)
{
  if (GP_OK != gp_camera_trigger_capture( camera_, context_ )) {
    logger_ << "gp_camera_trigger_capture() FAILED!!" << std::endl;
    return false;
  }

  assert(wait_timeout > 0);

  CameraEventType event_type;
  void *event_data;
  do {
    if (GP_OK == gp_camera_wait_for_event(camera_, wait_timeout, &event_type, &event_data, context_)) {
      if (event_type == GP_EVENT_FILE_ADDED) {
        CameraFilePath camera_file_path(*((CameraFilePath*) event_data));
#if 0
        printf("Pathname on the camera: %s/%s\n", camera_file_path.folder,
               camera_file_path.name);
#endif
        path = std::filesystem::path(camera_file_path.folder) / camera_file_path.name;
      }
    } else {
      logger_ << "gp_camera_wait_for_event()" << std::endl;
      return false;
    }
  } while (event_type != wait_for);

  return true;
}

bool do_download(const std::filesystem::path &src, const std::filesystem::path &dst, const bool rm_src /* = false */)
{
  int fd, error_code;
  CameraFile *photo_file;
  CameraFilePath photo_file_path;

  strcpy( photo_file_path.folder, src.parent_path().c_str());
  strcpy( photo_file_path.name, src.filename().c_str());

  fd = open( dst.c_str(), O_CREAT|O_WRONLY, 0644 );
  error_code = gp_file_new_from_fd( &photo_file, fd );
  if (error_code < GP_OK) {
    logger_ << "gp_file_new_from_fd() FAILED!!" << std::endl;
    gp_context_error( context_, "Could not create a new image file from %s%s (error code %d)\n", photo_file_path.folder, photo_file_path.name, error_code );
    gp_file_free( photo_file );
    return false;
  }

  error_code = gp_camera_file_get( camera_, photo_file_path.folder, photo_file_path.name, GP_FILE_TYPE_NORMAL, photo_file, context_ );
  if (error_code < GP_OK) {
    logger_ << "gp_camera_file_get() FAILED!!" << std::endl;
    gp_context_error( context_, "Could not get file %s%s (error code %d)\n", photo_file_path.folder, photo_file_path.name, error_code );
    gp_file_free( photo_file );
    return false;
  }

  if (rm_src) {
    error_code = gp_camera_file_delete( camera_, photo_file_path.folder, photo_file_path.name, context_ );
    if( error_code < GP_OK )
      {
        logger_ << "gp_camera_file_delete() FAILED!!" << std::endl;
        gp_context_error( context_, "Could not delete file %s%s  (error code %d)\n", photo_file_path.folder, photo_file_path.name, error_code );
        gp_file_free( photo_file );
        return false;
      }
  }

  gp_file_free( photo_file );
  return true;
}

bool download_file(const std::filesystem::path &src, const std::filesystem::path &dst, const bool lock = false)
{
  // ROS_DEBUG_STREAM("downloading: " << src << " --> " << dst);
  logger_ << "downloading: " << src << " --> " << dst << std::endl;
  if (lock) photo_mutex_.lock();
  auto r = do_download(src, dst, delete_image_after_download_);
  if (lock) photo_mutex_.unlock();
  if (!r) {
    // ROS_WARN_STREAM("failed to download: " << src);
    logger_ << "failed to download: " << src << std::endl;
  }
  return r;
}

void trigger()
{
  std::filesystem::path path;           /**< path on camera */
  photo_mutex_.lock();
  // latest_trigger_time_ = ros::Time::now();
  if (do_trigger(path, trigger_wait_timeout_ * 1000)) {
    // publish_image_info(path);
    switch (download_image_) {
    case 0:
      // do nothing
      break;
    case 1:
      download_file(path, download_path_ / path.filename());
      // download_file(path, download_path_ / path.filename(), true);
      break;
    case 2:
      // make_download_thread(path, download_path_ / path.filename());
      break;
    }

    logger_ << "got an image:" << path << std::endl;
  }
  photo_mutex_.unlock();
}

int main(int argc, char **argv)
{
  if (gp_camera_new(&camera_) != GP_OK) {
    logger_ << "gp_camera_new() FAILED!" << std::endl;
    return -1;
  }
  context_ = gp_context_new();
  std::cout << "starting..." << std::endl;

  while(true) trigger();

  return 0;
}
